"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.handler = exports.builder = exports.desc = exports.command = void 0;

var _graphql = require("graphql");

var _papaparse = _interopRequireDefault(require("papaparse"));

var _fsExtra = _interopRequireDefault(require("fs-extra"));

var _path = _interopRequireDefault(require("path"));

var _chalk = _interopRequireDefault(require("chalk"));

var _mkdirp = _interopRequireDefault(require("mkdirp"));

var _md5Base = _interopRequireDefault(require("md5-base64"));

var _moment = _interopRequireDefault(require("moment"));

var _util = require("./util");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

const DEFAULT_BATCH_SIZE = 1000000;
/**
 * Plugin boilerplate
 */

const command = 'mload <fileOrDir>';
exports.command = command;
const desc = 'Load a file or directory tree into Maana Q';
exports.desc = desc;
const builder = {
  mutation: {
    alias: 'm',
    description: 'mutation to call'
  },
  type: {
    alias: 't',
    description: 'type of entities (e.g., Person)'
  },
  ndfout: {
    alias: 'n',
    description: 'directory to store NDF conversion of input'
  },
  endpoint: {
    alias: 'e',
    description: 'endpoint to use'
  },
  batchsize: {
    alias: 'b',
    description: 'max records to send at once'
  }
};
/**
 * A collection of the different results while processing and uploading the
 * files so we can build a report for the user at the end of the command.
 */

exports.builder = builder;
const fileResults = {
  ndfGenCnt: 0,
  // # of "generations" (based on file flushes)
  ndfFileCnt: 0,
  // how many NDF files were written
  succeed: 0,
  total: 0,
  errors: {
    mutation: [],
    dataRead: [],
    uploading: {},
    ndf: []
  },
  warnings: {
    typeCoercion: []
  }
};
/**
 * NDF ids are CHAR(25), so we translate all original ID references to conform
 */

const mkNdfId = function (id) {
  return id.length > 25 ? (0, _md5Base.default)(id).slice(0, -2) : id;
};
/**
 *
 * @param {*} type
 */


const typeIsList = function (type) {
  let isList = false;

  if (type.name && type.name.endsWith('Connection')) {
    isList = true;
  }

  while (!isList && ((0, _graphql.isListType)(type) || (0, _graphql.isNonNullType)(type) || type.kind === 'NON_NULL' || type.kind === 'LIST')) {
    if ((0, _graphql.isListType)(type) || type.kind === 'LIST') {
      isList = true;
      break;
    }

    type = type.ofType;
  }

  return isList;
};
/**
 * Read and parse a CSV file
 *
 * @param {*} context
 * @param {Path} parsedPath
 */


const readCsvFile = function (context, parsedPath) {
  const filePath = _path.default.format(parsedPath);

  context.spinner.start(`Parsing CSV file ${_chalk.default.yellow(filePath)}`);

  try {
    const text = (0, _util.readFile)(filePath); // Parse the whole file (may need to be smarter about this in the future...)

    const parsedCsv = _papaparse.default.parse(text, {
      header: true,
      skipEmptyLines: true
    });

    if (parsedCsv.meta.aborted) {
      throw new Error(`Error parsing CSV file ${filePath} rows: ${parsedCsv.data.length} errors: ${JSON.stringify(parsedCsv.errors)} meta: ${JSON.stringify(parsedCsv.meta)}`);
    }

    context.spinner.succeed(_chalk.default.green(`Parsed ${_chalk.default.yellow(filePath)} entities: ${_chalk.default.yellow(parsedCsv.data.length)} meta: ${_chalk.default.yellow(JSON.stringify(parsedCsv.meta))}`));

    if (parsedCsv.errors && parsedCsv.errors.length) {
      console.log(_chalk.default.red(`✘ Parse errors (${_chalk.default.yellow(parsedCsv.errors.length)}): ${(0, _util.ellipse)(JSON.stringify(parsedCsv.errors, null, 2), 500)}`));
    }

    return parsedCsv.data;
  } catch (error) {
    context.spinner.fail(_chalk.default.red(error));
  }
};
/**
 * Read and parse a JSON file
 *
 * @param {*} context
 * @param {Path} parsedPath
 */


const readJsonFile = function (context, parsedPath) {
  const filePath = _path.default.format(parsedPath);

  context.spinner.start(`Parsing JSON file ${_chalk.default.yellow(filePath)}`);

  try {
    const json = (0, _util.readJson)(filePath);
    const entityCnt = Array.isArray(json) ? json.length : 1;
    context.spinner.succeed(_chalk.default.green(`Done parsing JSON file ${_chalk.default.yellow(filePath)} entities: ${_chalk.default.yellow(entityCnt)}`));
    if (entityCnt === 1 && !Array.isArray(json)) return [json];
    return json;
  } catch (error) {
    context.spinner.fail(_chalk.default.red(error));
  }
};
/**
 * Read ALL of the data into memory
 * @@TODO: streaming interface
 *
 * @param {*} context
 * @param {Path} parsedPath
 */


const readData = function (context, parsedPath) {
  const ext = parsedPath.ext;

  switch (ext) {
    case '.csv':
    case '.tsv':
    case '.dat':
      return readCsvFile(context, parsedPath);

    case '.json':
      return readJsonFile(context, parsedPath);

    default:
      return null;
    // should be unreachable
  }
};
/**
 * Type cache to speed lookups
 */


const typeCache = {};
/**
 * Get the GraphQL field definition from a type
 *
 * @param {*} type
 * @param {String} fieldName
 */

const getField = function (type, fieldName) {
  const key = `${type.name}.${fieldName}`; // console.log("key", key);

  let entry = typeCache[key];

  if (!entry) {
    const fields = type.getFields();
    const field = fields ? fields[fieldName] : undefined;

    if (!field) {
      const msg = `Undefined field: ${_chalk.default.yellow(fieldName)}`;
      console.log(_chalk.default.red(`✘ ${msg}`));
      throw msg;
    } // console.log("field", field);
    // console.log("Field:", key);


    const namedType = (0, _graphql.getNamedType)(field.type); // console.log("- namedType:", namedType.name);

    const isList = typeIsList(field.type); // console.log("- isList:", isList);

    const isNonNull = (0, _graphql.isNonNullType)(field.type); // console.log("- isNonNull:", isNonNull);

    const isObject = (0, _graphql.isObjectType)(namedType); // console.log("- isObject:", isObject);

    entry = {
      field,
      isList,
      isNonNull,
      isObject,
      namedType
    };
    typeCache[key] = entry;
  }

  return entry;
};
/**
 * Get the named mutation field from the GraphQL schema
 *
 * @param {*} schema
 * @param {*} mutationName
 */


const getMutationField = function (schema, mutationName) {
  const mutationType = schema.getMutationType();

  if (!mutationType) {
    return console.log(_chalk.default.red(`✘ No mutation type in schema.`));
  }

  const fields = mutationType.getFields();
  const mutationField = fields[mutationName];

  if (!mutationField) {
    return console.log(_chalk.default.red(`✘ Mutation for "${_chalk.default.yellow(mutationName)}" not found.`));
  }

  console.log(`Using mutation ${_chalk.default.yellow(mutationField.name)}: ${_chalk.default.yellow(mutationField.description || '(no description)')}`);
  return mutationField;
};
/**
 * Get the input (argument) type for the GraphQL mutation
 *
 * @param {*} schema
 * @param {*} mutationField
 */


const getInputTypeDef = function (schema, mutationField) {
  // Find the 'input' argument, if any
  const inputs = mutationField.args.filter(function (x) {
    return x.name === 'input';
  });

  if (!inputs || inputs.length !== 1) {
    return console.log(_chalk.default.red(`✘ Input argument missing for ${_chalk.default.yellow(mutationField.name)}.`));
  }

  const inputType = inputs[0].type; // Get the underlying type name (i.e., strip '!' and '[]')

  const namedType = (0, _graphql.getNamedType)(inputs[0].type); // Get the actual type definition from the schema

  const typeDef = schema.getType(namedType);
  console.log(`Input type ${_chalk.default.yellow(typeDef.name)}: ${_chalk.default.yellow(typeDef.description || '(no description)')}`);
  return {
    inputType,
    typeDef
  };
};
/**
 * Coerce the input type to match the output type
 *
 * @param {*} type
 * @param {*} val
 * @param {*} def
 */


const coerce = function ({
  type,
  val,
  def = null,
  quoted = false,
  isoDate = false
}) {
  let rval = def;

  if (type == 'Float') {
    if (typeof val == 'string') {
      if (val && val.trim()) {
        rval = Number(val);

        if (isNaN(rval)) {
          const message = _chalk.default.yellow(`☛ WARNING: Was not able to coerce Float from String--"${_chalk.default.red(val)}": Not a number (NaN). Instead used default value: ${def}`);

          console.log(message);
          fileResults.warnings.typeCoercion.push(message);
          rval = def;
        }
      } else {
        const message = _chalk.default.yellow(`☛ WARNING: Was not able to coerce Float from null or empty string input. Instead used default value: ${def}`);

        console.log(message);
        fileResults.warnings.typeCoercion.push(message);
      }
    } else if (typeof val == 'number') {
      rval = val;
    }
  } else if (type == 'Int') {
    if (typeof val == 'string') {
      if (val && val.trim()) {
        rval = parseInt(Number(val));

        if (isNaN(rval)) {
          const message = _chalk.default.yellow(`☛ WARNING: Was not able to coerce Int from String--"${_chalk.default.red(val)}": Not a number (NaN). Instead used default value: ${def}`);

          console.log(message);
          fileResults.warnings.typeCoercion.push(message);
          rval = def;
        }
      } else {
        const message = _chalk.default.yellow(`☛ WARNING: Was not able to coerce Int from null or empty string input. Instead used default value: ${def}`);

        console.log(message);
        fileResults.warnings.typeCoercion.push(message);
      }
    } else if (typeof val == 'number') {
      rval = val;
    }
  } else if (type == 'Boolean') {
    if (typeof val == 'string') {
      const lcVal = val.toLowerCase();

      if (lcVal == 'true') {
        rval = true;
      } else if (lcVal == 'false') {
        rval = false;
      }
    } else if (typeof val == 'boolean') {
      rval = val;
    }
  } else if (type == 'Date' || type == 'DateTime' || type == 'Time') {
    const cvtDate = function () {
      let dateStr;

      if (type == 'Time') {
        const date = (0, _util.parseTime)(val);
        if (isNaN(date)) throw new Error(`Invalid time: ${val}`);
        dateStr = val;
      } else {
        const date = (0, _moment.default)(val);
        if (!date.isValid()) throw new Error(`Invalid date/time: ${val}`);

        if (type == 'Date' && isoDate || type == 'DateTime') {
          dateStr = date.toISOString();
        } else if (type == 'Date') {
          dateStr = date.format('YYYY-MM-DD');
        }
      }

      if (quoted) dateStr = `"${dateStr}"`;
      return dateStr;
    };

    rval = !val || val == '' ? def : cvtDate();
  } else if (type == 'JSON') {
    try {
      if (!val) {
        rval = def;
      } else {
        // Strings are read from CSV files with escaping, i.e. string
        //   {\"id\":\"1234\"}
        // will be stored internally as
        //   "{\\\"id\\\":\\\"1234\\\"}"
        // Replacing values and parsing it will produce an object
        // Stringifying it first time will produce valid JSON
        // And stringifying it second time will escape quotes etc
        rval = JSON.stringify(JSON.stringify(JSON.parse(val.replace(/\\\"/g, '"'))));
      }
    } catch (error) {
      throw new Error(`Invalid JSON value is provided: ${val}. Error: ${error}`);
    }
  } else if (!quoted && typeof val == 'string') {
    rval = val;
  } else {
    rval = JSON.stringify(val);
  } // console.log("type", type, "rval", rval, "quoted", quoted);


  return rval;
};
/**
 * Generate the full mutation to upload based on the data
 *
 * @param {*} mutationField
 * @param {*} inputType
 * @param {*} data
 */


const buildMutation = function (mutationField, inputType, typeDef, data) {
  // console.log("data", data);
  const instances = data.map(function (row) {
    try {
      const fields = Object.keys(row).map(function (fieldName) {
        let value = row[fieldName];
        const {
          field,
          isList,
          isNonNull,
          namedType
        } = getField(typeDef, fieldName); // Parse a list

        const valueType = typeof value;

        if (isList && valueType === 'string') {
          value = value.split(',');
        }

        if (Array.isArray(value)) {
          if (!isList) {
            const msg = `Unexpected collection for ${_chalk.default.yellow(fieldName)}: ${_chalk.default.yellow(field.type)} => ${_chalk.default.yellow(JSON.stringify(value))}`;
            console.log(_chalk.default.red(`✘ ${msg}`));
            throw msg;
          }

          if (value.length === 0) {
            return `${fieldName}: null`;
          }

          const values = value.map(function (v) {
            return coerce({
              type: namedType,
              val: v,
              quoted: true
            });
          });
          return `${fieldName}: [${values.join(',')}]`;
        } else if (isList) {
          const msg = `Expected collection for ${_chalk.default.yellow(fieldName)}: ${field.type} => ${JSON.stringify(value)}`;
          console.log(_chalk.default.red(`✘ ${msg}`));
          throw msg;
        }

        return `${fieldName}: ${coerce({
          type: namedType,
          val: value,
          quoted: true
        })}\n`;
      }).join(',');
      return `{${fields}}`;
    } catch (error) {
      console.log(_chalk.default.red(`✘ ${error}`));
    }
  }); // Handle list input type

  if (typeIsList(inputType)) {
    const list = instances.join(', ');
    const mutation = `mutation { ${mutationField.name}(input:[${list}]) }`;
    return mutation;
  } // Single input type


  const mutations = instances.map(function (instance, idx) {
    return `m${idx}: ${mutationField.name}(input:${instance})`;
  }).join('\n');
  return `mutation { ${mutations} }`;
};
/**
 * File load dispatcher: NDF conversion or upload via mutation
 *
 * @param {*} context
 * @param {*} filePath
 */


const load = /*#__PURE__*/function () {
  var _ref = _asyncToGenerator(function* (context, filePath) {
    // console.log("filePath", filePath);
    // Parse the path into its components
    const parsedPath = _path.default.parse(filePath); // console.log("parsedPath", parsedPath);
    // Only accept supported file types


    const ext = parsedPath.ext.toLowerCase();

    switch (ext) {
      case '.csv':
      case '.tsv':
      case '.dat':
        break;

      case '.json':
        break;

      default:
        // console.log(chalk.yellow(`skipping unsupported file type: ${filePath}`));
        return;
    }

    fileResults.total++; // Dispatch to the right operation:
    // - convert input to NDF
    // - upload entities via mutation
    // The simple test is: if an NDF output directory was specified, then we
    // are performing a conversion, otherwise we are uploading via mutation

    const ndfout = context.argv.ndfout;
    if (ndfout) return convertToNdf(context, parsedPath);
    return uploadViaMutation(context, parsedPath);
  });

  return function load(_x, _x2) {
    return _ref.apply(this, arguments);
  };
}();
/**
 * NDF conversion
 */


const convertToNdf = /*#__PURE__*/function () {
  var _ref2 = _asyncToGenerator(function* (context, parsedPath) {
    // Construct string version of file path
    const filePath = _path.default.format(parsedPath);

    context.spinner.start(`Converting ${_chalk.default.yellow(filePath)}`); // Get the NDF output directory

    const ndfOut = context.argv.ndfout;
    const ndfPath = _path.default.isAbsolute(ndfOut) ? ndfOut : _path.default.resolve(process.cwd(), ndfOut); // console.log('ndfPath', ndfPath)
    // Infer the typenaame

    const typeName = context.argv.type || parsedPath.name; // console.log('typeName', typeName)
    // Get the GraphQL type definition

    const baseType = context.schema.getType(typeName);

    if (!baseType) {
      fileResults.errors.ndf.push({
        file: filePath,
        msg: `type not found in schema: ${typeName}`
      });
      return;
    } // console.log(
    //   `Base type ${chalk.yellow(baseType.name)}: ${chalk.yellow(
    //     baseType.description || '(no description)'
    //   )}`
    // )
    // Ensure the type has an id field


    const idFieldInfo = getField(baseType, 'id');

    if (!idFieldInfo) {
      fileResults.errors.ndf.push({
        file: filePath,
        msg: `type has no id field: ${typeName}`
      });
      return;
    } // Read the data
    // @@TODO: streaming


    const data = readData(context, parsedPath);

    if (!data) {
      fileResults.errors.dataRead.push({
        file: filePath
      });
      return;
    } // console.log('data', data)
    // Build internal state (resettable) for the different NDF value types


    let nodes = [];
    let lists = [];
    let relations = [];

    const flush = function () {
      // Any output to produce?
      if ((0, _util.isNullOrEmpty)(nodes) && (0, _util.isNullOrEmpty)(lists) && (0, _util.isNullOrEmpty)(relations)) return;
      fileResults.ndfGenCnt++;
      const outName = `${fileResults.ndfGenCnt}`.padStart(5, '0'); // NOTE: the file paths must exist, even if they are empty
      //       or "prisma --import" will throw an exception

      const nodePath = ensureDir(ndfPath, 'nodes');

      if (!(0, _util.isNullOrEmpty)(nodes)) {
        writeNdfFile(nodePath, 'nodes', outName, nodes);
        nodes = [];
        fileResults.ndfFileCnt++;
      }

      const listPath = ensureDir(ndfPath, 'lists');

      if (!(0, _util.isNullOrEmpty)(lists)) {
        writeNdfFile(listPath, 'lists', outName, lists);
        lists = [];
        fileResults.ndfFileCnt++;
      }

      const relPath = ensureDir(ndfPath, 'relations');

      if (!(0, _util.isNullOrEmpty)(relations)) {
        writeNdfFile(relPath, 'relations', outName, relations);
        relations = [];
        fileResults.ndfFileCnt++;
      }
    }; // Stats


    let entityCnt = 0;
    let errorCnt = 0; // Don't allow duplicate records

    const dedupe = {};
    data.forEach(function (entity) {
      // console.log('entity', entity)
      if (!entity.id) {
        fileResults.errors.ndf.push({
          file: filePath,
          msg: `entity missing id: ${JSON.stringify(entity)}`
        });
        errorCnt++;
        return;
      } // Make a NDF-compliant ID


      const id = mkNdfId(entity.id); // Dedupe

      if (dedupe[id]) {
        fileResults.errors.ndf.push({
          file: filePath,
          msg: `duplicate id: ${entity.id} => ${dedupe[id]}`
        });
        errorCnt++;
        return;
      }

      dedupe[id] = entity.id; // Intermediate state for this entity

      const nodeValues = {};
      const listValues = {};
      const relValues = {}; // Process all the fields

      Object.keys(entity).filter(function (x) {
        return x != 'id';
      }).forEach(function (fieldName) {
        const fieldRes = getField(baseType, fieldName); // console.log('field', fieldName, fieldRes)

        const {
          field,
          isList,
          isNonNull,
          isObject,
          namedType
        } = fieldRes; // What type of field is it?

        if (isList) {
          if (isObject) {
            // add a list of relations
            // add a list of values
            relValues[fieldName] = {
              toType: namedType.name,
              toIds: entity[fieldName].map(mkNdfId)
            };
          } else {
            // add a list of values
            listValues[fieldName] = namedType.name === 'ID' ? mkNdfId(entity[fieldName]) : entity[fieldName];
          }
        } else if (isObject) {
          // add a relation
          relValues[fieldName] = {
            toType: namedType.name,
            toIds: [mkNdfId(entity[fieldName])]
          }; // always a collection
        } else {
          // add a value
          const value = namedType.name === 'ID' ? mkNdfId(entity[fieldName]) : coerce({
            type: namedType.name,
            val: entity[fieldName],
            isoDate: true
          });

          if (value != null && value != undefined) {
            nodeValues[fieldName] = value;
          }
        }
      }); // Generate output for this entity
      // ... add the set of leaf field values

      nodes.push({
        _typeName: typeName,
        id,
        ...nodeValues
      });

      if (!(0, _util.isNullOrEmpty)(listValues)) {
        // add the set of list field values (i.e., lists)
        lists.push({
          _typeName: typeName,
          id,
          ...listValues
        });
      }

      if (!(0, _util.isNullOrEmpty)(relValues)) {
        // add the relationship pairs
        Object.keys(relValues).forEach(function (fieldName) {
          const relValue = relValues[fieldName];
          relValue.toIds.forEach(function (toId) {
            relations.push([{
              _typeName: typeName,
              id,
              fieldName
            }, {
              _typeName: relValue.toType,
              id: toId
            }]);
          });
        });
      }

      entityCnt++;
      if (entityCnt % 10000 === 0) flush();
    }); // entity loop

    flush();

    if (!errorCnt) {
      fileResults.succeed++;
      context.spinner.succeed(_chalk.default.green(`Converted ${_chalk.default.yellow(filePath)} without errors`));
    } else {
      context.spinner.fail(_chalk.default.red(`✘ Converted ${_chalk.default.yellow(filePath)} with ${errorCnt} errors`));
    }
  });

  return function convertToNdf(_x3, _x4) {
    return _ref2.apply(this, arguments);
  };
}();
/**
 * Given a root and directory, combine them and ensure the full path exists
 *
 * @param {*} root
 * @param {*} dir
 */


const ensureDir = function (root, dir) {
  const outDir = _path.default.resolve(root, dir); // console.log("outdir", outDir);
  // Ensure the output path exists


  _mkdirp.default.sync(outDir);

  return outDir;
};
/**
 * Write data to one of the NDF file types
 *
 * @param {*} ndfPath
 * @param {*} valueType
 * @param {*} typeName
 * @param {*} values
 */


const writeNdfFile = function (basePath, valueType, typeName, values) {
  const outName = `${typeName}.json`; // console.log("outName", outName);

  const outPath = _path.default.resolve(basePath, outName); // console.log("outpath", outPath);


  const data = JSON.stringify({
    valueType,
    values
  }, null, 2);

  _fsExtra.default.writeFileSync(outPath, data, 'utf8');
};
/**
 * Upload entities via mutation
 */


const uploadViaMutation = /*#__PURE__*/function () {
  var _ref3 = _asyncToGenerator(function* (context, parsedPath) {
    // Generate a mutation name from base name
    const genMutationName = function (baseName) {
      return `add${baseName}s`;
    }; // Infer the mutation


    const mutationName = context.argv.mutation || genMutationName(parsedPath.name); // console.log("useMutationName", useMutationName);
    // Construct string version of file path

    const filePath = _path.default.format(parsedPath);

    const mutationField = getMutationField(context.schema, mutationName);

    if (!mutationField) {
      fileResults.errors.mutation.push({
        file: filePath,
        mutation: mutationName
      });
      return;
    } // console.log("mutationField", mutationField);


    const {
      inputType,
      typeDef
    } = getInputTypeDef(context.schema, mutationField);

    if (!inputType) {
      fileResults.errors.mutation.push({
        file: filePath,
        mutation: mutationName
      });
      return;
    } // console.log("inputType", inputType);
    // console.log("inputType fields", inputType.getFields());
    // Read the data
    // @@TODO: streaming


    const data = readData(context, parsedPath);

    if (!data) {
      fileResults.errors.dataRead.push({
        file: filePath
      });
      return;
    }

    let offset = 0;
    const batchSize = context.argv.batchsize || DEFAULT_BATCH_SIZE; // console.log("batchSize", batchSize);

    const totalLength = data.length;

    while (offset < totalLength) {
      let end = offset + batchSize;

      if (end > totalLength) {
        end = totalLength;
      }

      const batch = data.slice(offset, end);
      let mutation;

      try {
        mutation = buildMutation(mutationField, inputType, typeDef, batch);
      } catch (error) {
        fileResults.errors.mutation.push({
          file: filePath,
          mutation: mutationName
        });
        return;
      }

      if (!mutation) {
        fileResults.errors.mutation.push({
          file: filePath,
          mutation: mutationName
        });
        return;
      } // console.log('mutation', mutation)


      const incErrors = function () {
        if (!fileResults.errors.uploading[filePath]) {
          fileResults.errors.uploading[filePath] = 0;
        }

        fileResults.errors.uploading[filePath]++;
      };

      try {
        if (offset === 0 && totalLength > end || offset > 0) {
          context.spinner.start(_chalk.default.yellow(`Uploading batch ${offset}-${end} of ${totalLength}`));
        } else {
          context.spinner.start(_chalk.default.yellow(`Uploading`));
        }

        const result = yield context.client.request(mutation); // console.log("result", result);

        if (result['errors']) {
          context.spinner.fail(_chalk.default.red(`Call failed: ${JSON.stringify(result['errors'])}`));
          incErrors();
        } else {
          const nulls = Object.keys(result).filter(function (x) {
            return result[x] === null;
          });

          if (nulls.length === 0) {
            context.spinner.succeed(_chalk.default.green(`Call succeeded: ${_chalk.default.yellow((0, _util.ellipse)(JSON.stringify(result)))}`));
          } else {
            const total = Object.keys(result).length;
            context.spinner.succeed(_chalk.default.green(`${nulls.length} out of ${total} mutations changed nothing`));
          }
        }
      } catch (error) {
        let msg;

        if (typeof error === 'object') {
          const response = error.response;

          if (response.error) {
            msg = `GraphQL Error: ${response.error || '(none)'}, status: ${response.status}`;
          } else if (response.errors) {
            const errMsgs = response.errors.map(function (e) {
              return (0, _util.ellipse)(e.message, 2000);
            });

            if (errMsgs.length === 1) {
              msg = errMsgs[0];
            } else {
              msg = `[${errMsgs.join(',\n')}]`;
            }
          } else {
            msg = (0, _util.ellipse)(JSON.stringify(error), 2000);
          }
        } else if (typeof error === 'string') {
          msg = error;
        }

        context.spinner.fail(_chalk.default.red(`Exception: ${_chalk.default.yellow(msg)}`));
        incErrors();
      }

      offset += batchSize;
    }

    if (!fileResults.errors.uploading[filePath]) {
      fileResults.succeed++;
    }
  });

  return function uploadViaMutation(_x5, _x6) {
    return _ref3.apply(this, arguments);
  };
}();
/**
 * Main command handler
 *
 * @param {*} context
 * @param {*} argv
 */


const handler = /*#__PURE__*/function () {
  var _ref4 = _asyncToGenerator(function* (context, argv) {
    // Capture the arguments
    context.argv = argv; // console.log("argv", argv);
    // Get the project level configuration from .graphqlconfig

    const config = yield context.getProjectConfig(); // console.log("config", config);
    // Given the configuration and the arguments, what endpoint do we use?

    context.endpoint = (0, _util.getEndpoint)(config, argv);
    if (!context.endpoint) return; // Given the endpoint, what is the client we are using

    context.client = context.endpoint.getClient(); // This is the schema string
    // const schemaSDL = config.getSchemaSDL();
    // console.log("schemaSDL", schemaSDL);
    // Load the project's schema (AST)

    context.schema = (0, _util.getSchema)(config); // console.log("schema", schema);
    // const configPath = config.configPath;
    // console.log("configPath", configPath);
    // Load the project's extensions

    context.extensions = config.extensions || {}; // console.log("extensions", context.extensions);
    // Load the project's default Maana options

    context.options = context.extensions.maana || {}; // console.log("options", context.options);
    // What are we loading?

    const fileOrDir = argv.fileOrDir;

    if (!fileOrDir) {
      return console.log(_chalk.default.red(`✘ Must specify a file or directory`));
    } // Are we processing an entire folder?


    if (_fsExtra.default.lstatSync(fileOrDir).isDirectory()) {
      // Get the set of files
      const filePaths = (0, _util.getAllFiles)(fileOrDir); // Iterate SEQUENTIALLY over all the files (recursively) in the folder tree rooted at 'folder'

      for (let i = 0; i < filePaths.length; i++) {
        yield load(context, filePaths[i]);
      }
    } else {
      yield load(context, fileOrDir);
    } // output the results from processing and uploading the files


    buildReport();
  });

  return function handler(_x7, _x8) {
    return _ref4.apply(this, arguments);
  };
}();
/**
 * Generate a report of actions and errors
 */


exports.handler = handler;
const DIVIDER = '-------------------------------------------------------------------------------';

const buildReport = function () {
  // NDF conversion stats
  if (fileResults.ndfFileCnt > 0) {
    console.log(_chalk.default.green(DIVIDER));
    console.log('NDF Conversion:');
    console.log(_chalk.default.green(`✔ Files processed    : ${_chalk.default.yellow(`${fileResults.total}`)}` // assume we only did conversion
    ));
    console.log(_chalk.default.green(`✔ Generations        : ${_chalk.default.yellow(`${fileResults.ndfGenCnt}`)}`));
    console.log(_chalk.default.green(`✔ NDF files generated: ${_chalk.default.yellow(`${fileResults.ndfFileCnt}`)}`));
  } // let the user know how many files had no issues


  if (fileResults.succeed) {
    console.log(_chalk.default.green(DIVIDER));
    console.log(_chalk.default.green(`✔ ${fileResults.succeed} of ${fileResults.total} file(s) succeeded`));
  } // figure out how many errors we have


  const uploadErrorKeys = Object.keys(fileResults.errors.uploading);
  const numErrors = uploadErrorKeys.length + fileResults.errors.mutation.length + fileResults.errors.dataRead.length + fileResults.errors.ndf.length; // only show information about errors if there are any

  if (numErrors) {
    // deliver the sad news
    console.log(_chalk.default.red(DIVIDER)); // let the user know the total number of files that had issues

    console.log(_chalk.default.red(`✘ ${numErrors} issue(s):`)); // tell the user about NDF conversion issues

    if (fileResults.errors.ndf.length) {
      console.log(_chalk.default.red(`  ${fileResults.errors.ndf.length} NDF conversion`));
      fileResults.errors.ndf.forEach(function (e) {
        return console.log(`    ${_chalk.default.yellow(e.file)}: ${_chalk.default.red(e.msg)}`);
      });
    } // tell the user about issues with mutations


    if (fileResults.errors.mutation.length) {
      console.log(_chalk.default.red(`  ${fileResults.errors.mutation.length} file(s) had issues finding or creating the mutation`));
      fileResults.errors.mutation.forEach(function (e) {
        return console.log(`    ${_chalk.default.yellow('File')} ${e.file} ${_chalk.default.yellow('with mutation')} ${e.mutation}`);
      });
    } // tell the user about issues with reading the data off the disk


    if (fileResults.errors.dataRead.length) {
      console.log(_chalk.default.red(`  ${fileResults.errors.dataRead.length} file(s) had issues loading the data from disk`));
      fileResults.errors.dataRead.forEach(function (e) {
        return console.log(`    ${_chalk.default.yellow('File')} ${e.file}`);
      });
    } // tell the user about issues uploading the data


    if (uploadErrorKeys.length) {
      console.log(_chalk.default.red(`  ${uploadErrorKeys.length} file(s) had issues uploading data`));
      uploadErrorKeys.forEach(function (file) {
        return console.log(`    ${_chalk.default.yellow('File')} ${file} ${_chalk.default.yellow('in')} ${fileResults.errors.uploading[file]} ${_chalk.default.yellow('batch(es)')}`);
      });
    } // let the user know how to find more information about the issues


    console.log(_chalk.default.red(DIVIDER));
    console.log(_chalk.default.red('For more information on the errors, please look through the full output of the command'));
  } else {
    // let the user know that no issues happened during the command
    console.log(_chalk.default.green(DIVIDER));
    console.log(_chalk.default.green('✔ No errors'));
  } //-----------------------------WARNINGS---------------------------------
  //Let user know about warnings during file upload process


  const warnNum = fileResults.warnings.typeCoercion.length;

  if (warnNum) {
    console.log(_chalk.default.yellowBright(DIVIDER));
    console.log(_chalk.default.yellowBright(`☛ Warning summary: ${warnNum} warning(s) total`));

    if (fileResults.warnings.typeCoercion.length > 0) {
      console.log(_chalk.default.yellow(`☛ Type coercion warnings: ${fileResults.warnings.typeCoercion.length}`));
      fileResults.warnings.typeCoercion.forEach(function (warn) {
        console.log(warn);
      });
    }
  }
};