"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.handler = exports.builder = exports.desc = exports.command = void 0;

var _n = _interopRequireDefault(require("n3"));

var _fsExtra = _interopRequireDefault(require("fs-extra"));

var _path = _interopRequireDefault(require("path"));

var _mkdirp = _interopRequireDefault(require("mkdirp"));

var _chalk = _interopRequireDefault(require("chalk"));

var _url = _interopRequireDefault(require("url"));

var _util = require("./util");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

// Plugin boilerplate
const command = 'monto <inputPath>';
exports.command = command;
const desc = "Interpret an ontology, in the form of assertions (subject, predicate, object), into Maana Q's CKG (GraphQL types + Links & Relations).";
exports.desc = desc;
const builder = {
  inputPath: {
    description: 'Path to the input file.  Supported types:  Turtle, TriG, N-Triples, N-Quads, and Notation3 (N3)'
  },
  outputDir: {
    alias: 'o',
    description: 'Output directory (default = ./<inputPath dir>)'
  },
  map: {
    alias: 'm',
    description: 'Custom ontology map'
  }
}; // Default ontology map

exports.builder = builder;
const defOntoMap = {
  //
  // Assumed types
  //
  givenTypes: {
    'http://www.w3.org/2000/01/rdf-schema#Class': {},
    'http://www.w3.org/2002/07/owl#Thing': {},
    'http://www.w3.org/2000/01/rdf-schema#Resource': {}
  },
  //
  // Predicates
  //
  // "hasType" relation
  typeP: {
    'http://www.w3.org/1999/02/22-rdf-syntax-ns#type': {}
  },
  // "hasDomain" relation
  domainP: {
    'http://schema.org/domainIncludes': {},
    'http://www.w3.org/2000/01/rdf-schema#domain': {}
  },
  // "hasRange" relation
  rangeP: {
    'http://schema.org/rangeIncludes': {},
    'http://www.w3.org/2000/01/rdf-schema#range': {}
  },
  // "subClassOf" relation
  subclassP: {
    'http://www.w3.org/2000/01/rdf-schema#subClassOf': {}
  },
  // "comment" relation
  commentP: {
    'http://www.w3.org/2000/01/rdf-schema#comment': {}
  },
  // skip
  ignoreP: {
    'http://www.w3.org/2000/01/rdf-schema#label': {}
  },
  //
  // Objects
  //
  // "typeOf" Property
  propertyTypes: {
    'http://www.w3.org/1999/02/22-rdf-syntax-ns#Property': {}
  },
  // "typeOf" Class
  classTypes: {
    'http://www.w3.org/2000/01/rdf-schema#Class': {},
    'http://www.w3.org/2002/07/owl#Class': {}
  }
}; // Helper to track of self-identifying objects

const indexedItem = function (index, k) {
  return (0, _util.getOrSet)(index, k, {
    id: k
  });
}; // Helper to extract a name from a URI


const nameFromUri = function (uri) {
  if (!uri) return '#UNDEFINED#';

  const parsedUri = _url.default.parse(uri);

  return parsedUri.hash ? parsedUri.hash.slice(1) : parsedUri.path.split('/').slice(-1)[0];
};

const handler = /*#__PURE__*/function () {
  var _ref = _asyncToGenerator(function* (context, argv) {
    // Grab local copies of arguments
    const inputPathname = argv.inputPath; // console.log('inputFilename', inputFilename)
    // Parse the path into its components

    const inputPath = _path.default.parse(inputPathname); // console.log("inputPath", inputPath);
    // Generate output path and filenames based on input filename, if not provided


    let outputDir = argv.outputDir;

    if (!outputDir) {
      outputDir = _path.default.join(inputPath.root, inputPath.dir, inputPath.name);
    }

    console.log('outputDir', outputDir);
    (0, _mkdirp.default)(outputDir, function (err) {
      if (err) {
        console.log(_chalk.default.red(`Can't create output directory:${_chalk.default.yellow(err)}`));
      }
    }); // Load or use default ontology map

    let ontoMap = defOntoMap;
    const ontoMapFilename = argv.map;

    if (ontoMapFilename) {
      ontoMap = Object.assign(defOntoMap, (0, _util.readJson)(ontoMapFilename));
    } // console.log("ontoMap", ontoMap);
    // Update UI


    context.spinner.start(`Parsing ${_chalk.default.yellow(inputPathname)}\n`); // Create a parser and input stream

    const parser = _n.default.Parser();

    const readStream = _fsExtra.default.createReadStream(inputPathname); // State


    const types = {};
    const properties = {};
    const comments = {};
    const subclasses = {};
    const superclasses = {};
    const unions = {};
    const relations = {}; // Helper to ensure a property is valid

    const isValidProp = function (p) {
      const prop = properties[p];

      if (!prop) {
        console.log(_chalk.default.red(`Property not defined: ${_chalk.default.yellow(p)}`));
        return false;
      }

      if (!prop.domains || prop.domains.length === 0) {
        console.log(_chalk.default.red(`Property missing domain(s): ${_chalk.default.yellow(prop.id)}`));
        return false;
      }

      if (!prop.ranges || prop.ranges.length === 0) {
        console.log(_chalk.default.red(`Property missing range(s): ${_chalk.default.yellow(prop.id)}`));
        return false;
      }

      return true;
    }; // Recursively walk a superclass hierarchy to form an entire collection


    const allSuperclasses = function (root) {
      const set = new Set();

      const visit = function (cur) {
        const supers = superclasses[cur];
        if (!supers) return;
        supers.filter(function (x) {
          return !set.has(x);
        }).forEach(function (x) {
          set.add(x);
          visit(x);
        });
      };

      visit(root);
      return [...set];
    }; // Assumed types


    for (let t in ontoMap.givenTypes) indexedItem(types, t); // Parse the input file


    let assertionCnt = 0;
    parser.parse(readStream, function (error, quad, prefixes) {
      //
      // Parse errors
      //
      if (error) {
        context.spinner.fail(_chalk.default.red(`Parse failed: ${_chalk.default.yellow(JSON.stringify(error))}`));
        return;
      } //
      // Process a quad
      //


      if (quad) {
        assertionCnt++;
        const {
          subject,
          predicate,
          object
        } = quad; // Drive from the predicate

        if (predicate.termType === 'NamedNode') {
          if (ontoMap.typeP[predicate.value]) {
            //
            // subject hasType object
            //
            // is object a class or a property?
            if (ontoMap.classTypes[object.value]) {
              indexedItem(types, subject.value);
            } else if (ontoMap.propertyTypes[object.value]) {
              indexedItem(properties, subject.value);
            }
          } else if (ontoMap.domainP[predicate.value]) {
            //
            // property applies to class
            //
            const prop = indexedItem(properties, subject.value);
            (0, _util.append)(prop, 'domains', object.value);
          } else if (ontoMap.rangeP[predicate.value]) {
            //
            // property takes values of type
            //
            const prop = indexedItem(properties, subject.value);
            (0, _util.append)(prop, 'ranges', object.value);
          } else if (ontoMap.subclassP[predicate.value]) {
            //
            // subject subClassof object
            //
            const subclass = subject.value;
            const superclass = object.value; // update the set of superclasses for this subclass

            (0, _util.append)(superclasses, subclass, superclass); // update the set of subclasses for this superclass

            (0, _util.append)(subclasses, superclass, subclass);
          } else if (ontoMap.commentP[predicate.value]) {
            //
            // Add comment
            //
            (0, _util.append)(comments, subject.value, (0, _util.stripCRLF)(object.value));
          } else if (ontoMap.ignoreP[predicate.value]) {//
            // ignore this assertion
            //
          } else {
            //
            // otherwise, record an arbitrary relation
            //
            const rel = indexedItem(relations, predicate.value);
            (0, _util.append)(rel, 'links', {
              subject,
              object
            });
          }
        } else {
          console.log(_chalk.default.red(`Unknown predicate term type: ${_chalk.default.yellow(predicate.termType)}`));
        }

        return;
      } // process quad
      //
      // End of parsing
      //


      context.spinner.succeed(_chalk.default.green(`Parsed ${_chalk.default.yellow(assertionCnt)} assertions`));

      if (types.length === 0) {
        console.log(_chalk.default.red(`No types were found.`));
        return;
      }

      let outputPathname = `${outputDir}/model.gql`;
      context.spinner.start(`Generating ${_chalk.default.yellow(outputPathname)}\n`); // Assign each of the properties to their respective types (domains)

      Object.keys(properties).filter(isValidProp).forEach(function (p) {
        const prop = properties[p]; // Derive the name of the property from its URI

        const propName = nameFromUri(p); // For each domain this is a property of...

        prop.domains.forEach(function (t) {
          const type = types[t];

          if (!type) {
            console.log(_chalk.default.red(`Missing domain type: ${t}`));
            return;
          } // Infer the type (from the range or ranges)


          let propType;

          if (prop.ranges.length > 1) {
            // Construct a union
            propType = propName.charAt(0).toUpperCase() + propName.slice(1);
            const union = prop.ranges.map(nameFromUri);
            unions[propType] = union;
          } else {
            // Simple range
            propType = nameFromUri(prop.ranges[0]);
          } // Any comments?


          const propComments = comments[p]; // Associate with type

          type[p] = {
            name: propName,
            type: propType,
            comments: propComments
          };
        });
      }); // Create an output stream

      const writeStream = _fsExtra.default.createWriteStream(outputPathname); // For all the types...


      Object.keys(types).forEach(function (t) {
        let type = types[t];
        const typeName = nameFromUri(type.id); // Emit any type-level comments

        const typeComments = comments[type.id];

        if (typeComments && typeComments.length > 0) {
          typeComments.forEach(function (c) {
            writeStream.write(`# ${c}\n`);
          });
        } // Introduce the type


        writeStream.write(`type ${typeName} {\n`); // Inline helper to emit properties for a type

        const emitProperties = function (theType) {
          // Emit the original subject URI as a comment
          writeStream.write(`\n  #\n  # ${theType.id} properties\n  #\n`);
          const props = Object.keys(theType).filter(function (p) {
            return p != 'id';
          });

          if (props.length === 0) {
            writeStream.write(`  # (none)\n`);
          } else {
            props.forEach(function (p) {
              const prop = theType[p]; // Emit any property-level comments

              if (prop.comments && prop.comments.length > 0) {
                prop.comments.forEach(function (c) {
                  writeStream.write(`  # ${c}\n`);
                });
              } // Emit the property


              writeStream.write(`  ${prop.name}: ${prop.type}\n`);
            });
          }
        }; // Current type


        emitProperties(type); // Merge the properties from the superclass hierarchy, if any

        const supers = allSuperclasses(t);

        if (supers && supers.length > 0) {
          supers.forEach(function (s) {
            const supertype = types[s];

            if (!supertype) {
              console.log(_chalk.default.red(`Unknown superclass: ${_chalk.default.yellow(s)}`));
              return;
            }

            emitProperties(supertype);
          });
        } // Close the body of the type definition


        writeStream.write(`}\n\n`);
      }); // Emit the unions

      for (let u in unions) {
        const union = unions[u];
        writeStream.write(`union ${u} = ${union.join(' | ')}\n`);
      } // close the stream


      writeStream.end(); // Generate the relations

      const relOut = Object.keys(relations).map(function (r) {
        return {
          id: r,
          name: nameFromUri(r)
        };
      });
      outputPathname = `${outputDir}/relation.json`;

      _fsExtra.default.writeFile(outputPathname, JSON.stringify(relOut, null, 2)); // Generate the links


      const linkOut = [];

      for (let r in relations) {
        const rel = relations[r];

        for (let l of rel.links) {
          linkOut.push({
            rel: r,
            from: l.subject.value,
            to: l.object.value
          });
        }
      }

      outputPathname = `${outputDir}/link.json`;

      _fsExtra.default.writeFile(outputPathname, JSON.stringify(linkOut, null, 2)); //
      // All done
      //


      context.spinner.succeed(_chalk.default.green(`Generated ${_chalk.default.yellow(Object.keys(types).length)} types, ${_chalk.default.yellow(Object.keys(unions).length)} unions, ${_chalk.default.yellow(relOut.length)} relations, and ${_chalk.default.yellow(linkOut.length)} links`));
    });
  });

  return function handler(_x, _x2) {
    return _ref.apply(this, arguments);
  };
}();

exports.handler = handler;