"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.addHeadersToConfig = addHeadersToConfig;
exports.getSchema = exports.getEndpoint = exports.parseJson = exports.readJson = exports.readFile = exports.getAllFiles = exports.isNullOrEmpty = exports.getOrSet = exports.append = exports.parseTime = exports.ellipse = exports.stripCRLF = exports.capitalize = void 0;

var _fsExtra = _interopRequireDefault(require("fs-extra"));

var _path = _interopRequireDefault(require("path"));

var _chalk = _interopRequireDefault(require("chalk"));

var _graphql = require("graphql");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

//
// String utilities
//
// Ensure initial letter is capitalized
const capitalize = function (word) {
  return word[0].toUpperCase() + word.substr(1);
}; // Strip newlines from string


exports.capitalize = capitalize;

const stripCRLF = function (s) {
  return s.replace(/\r?\n|\r/g, '');
}; // Truncate text with an ellipse


exports.stripCRLF = stripCRLF;

const ellipse = function (str, max = 200) {
  if (str.length > max) {
    return str.substring(0, max) + '...';
  }

  return str;
};

exports.ellipse = ellipse;

const parseTime = function (time) {
  const currentDateString = new Date().toISOString();
  return new Date(currentDateString.substr(0, currentDateString.indexOf('T') + 1) + time);
};
/**
 * Update the set of values for this indexed collection
 */


exports.parseTime = parseTime;

const append = function (obj, k, v) {
  let e = obj[k];
  if (!e || e.length === 0) obj[k] = e = [];
  e.push(v);
  return e;
}; //
// Object utilities
//

/**
 * Ensure an entry exists in a dictionary
 */


exports.append = append;

const getOrSet = function (obj, k, v) {
  let e = obj[k];
  if (!e) obj[k] = e = v;
  return e;
};
/**
 * Emptiness test
 */


exports.getOrSet = getOrSet;

const isNullOrEmpty = function (obj) {
  if (obj === null || obj === undefined) return true;
  const type = typeof obj;

  if (type === 'object') {
    if (Array.isArray(obj)) {
      return obj.length === 0;
    }

    return Object.keys(obj).length === 0;
  }

  if (type === 'string') {
    return obj === '';
  }

  return false;
}; //
// File utilities
//

/**
 * Find all files inside a dir, recursively.
 * @param  {string} dir Dir path string.
 * @return {string[]} Array with all file names that are inside the directory.
 */


exports.isNullOrEmpty = isNullOrEmpty;

const getAllFiles = function (dir) {
  return _fsExtra.default.readdirSync(dir).reduce(function (files, file) {
    const name = _path.default.join(dir, file);

    const isDirectory = _fsExtra.default.statSync(name).isDirectory();

    return isDirectory ? [...files, ...getAllFiles(name)] : [...files, name];
  }, []);
};
/**
 * Synchronously read an entire UTF-8 encoded file, stripping initial BOM
 * code, if present
 * @param {file} file
 * @return {text}
 */


exports.getAllFiles = getAllFiles;

const readFile = function (file) {
  if (!file) {
    throw new Error('No file specifed for read');
  }

  let text = _fsExtra.default.readFileSync(file, 'utf-8');

  if (!text) {
    throw new Error(`No text in file: ${file}`);
  } // Catches EFBBBF (UTF-8 BOM) because the buffer-to-string
  // conversion translates it to FEFF (UTF-16 BOM)


  if (text.charCodeAt(0) === 0xfeff) {
    text = text.slice(1);
  } // console.log(chalk.green(`Read file: ${file} size: ${text.length}`));


  return text;
};

exports.readFile = readFile;

const readJson = function (file) {
  return parseJson(readFile(file));
};

exports.readJson = readJson;

const parseJson = function (str) {
  try {
    return JSON.parse(str);
  } catch (e) {
    throw new Error(`Error parsing ${ellipse(str)} as JSON: ${e}`);
  }
}; //
// GraphQL communication utilities
//


exports.parseJson = parseJson;

const getEndpoint = function (config, argv) {
  const extensions = config.extensions || {};
  const endpoints = extensions.endpoints || {};
  const key = argv.endpoint || Object.keys(endpoints)[0];

  if (!key) {
    return console.log(_chalk.default.red(`No endpoint found.`));
  }

  var endpoint = config.endpointsExtension.getEndpoint(key);

  if (!endpoint) {
    return console.log(_chalk.default.red(`No endpoint ${key} found.`));
  }

  if (typeof endpoint === 'string') {
    endpoint = {
      url: endpoint
    };
  }

  console.log(_chalk.default.green(`Using endpoint ${key}: ${endpoint.url}`));
  return endpoint;
};
/**
 * Adds the headers used in requests to the Maana API
 *
 * @param {Object} config to update
 */


exports.getEndpoint = getEndpoint;

function addHeadersToConfig(config) {
  if (config.projects) {
    Object.values(config.projects).forEach(addHeadersToConfig);
    return;
  }

  if (config.extensions && config.extensions.endpoints) {
    const endpoints = config.extensions.endpoints;
    Object.keys(endpoints).forEach(function (name) {
      let result = {};

      if (typeof endpoints[name] === 'string') {
        result.url = endpoints[name];
        result.headers = {};
      } else {
        result = endpoints[name];

        if (!result.headers) {
          result.headers = {};
        }
      }

      result.headers.Authorization = 'Bearer ${env:MAANA_AUTH_TOKEN}';
      endpoints[name] = result;
    });
  }
}
/**
 * Get the GraphQL schema for the project
 *
 * @param {*} config
 */


const getSchema = function (config) {
  const schemaPath = config.schemaPath;

  const schemaContents = _fsExtra.default.readFileSync(schemaPath).toString();

  return (0, _graphql.buildASTSchema)((0, _graphql.parse)(schemaContents));
};

exports.getSchema = getSchema;